-----------------------------------------------------------
-- Einfuehrung in die Informatik II                      --
-----------------------------------------------------------
-- Abgabe von                                            --
--   Artur-Ulrich Bachner                         --
--   Kerstin Eckart                               --
--   Juern-Jochen Laun                            --
-----------------------------------------------------------
-- Blatt 6, Aufgabe 1                                    --
-----------------------------------------------------------

with Ada.Text_Io, Ada.Integer_Text_Io, Ada.Numerics.Discrete_Random, LList;
use Ada.Text_Io, Ada.Integer_Text_Io;

procedure List_Cycle is
   -- content type
   type Content is new Natural;
   -- length of list
   type List_Index is new Natural range 1..100_000;

   package Natural_List is new LList(Content); use Natural_List;
   
   -- finds out whether a given list has a cycle
   -- see README.TXT for details
   function Has_Cycle(L: in List) return Boolean is
      Hare, Hedgehog: List := L;
   begin
      loop
         -- has the hare reached the end of the list?
         if ((Hare.Next = null) or else (Hare.Next.Next = null)) then
            return False;
         end if;
         -- or will he overtake the hedgehog in the next jump?
         if ((Hare.Next = Hedgehog) or else (Hare.Next.Next = Hedgehog)) then
            return True;
         end if;
         -- hare jumps
         Hare := Hare.Next.Next;
         -- hedgehog walks
         Hedgehog := Hedgehog.Next;
      end loop;
   end;
   
   -------------------------------------------------
   --b)
   -------------------------------------------------
   --identifies last cycle element and makes its
   --Next-pointer null
   function No_more_Cycle(L: in List) return List is
      Hare, Hedgehog, Wife: List := L;
      found: boolean := false; --Hare hasn't overtaken Hedgehog yet
   begin
      while not found loop
         --will the hare overtake the hedgehog in the next jump?
         if ((Hare.Next = Hedgehog) or else (Hare.Next.Next = Hedgehog)) then
            Hare:=Hedgehog;
            found:=true;
         else
            -- hare jumps
            Hare := Hare.Next.Next;
            -- hedgehog walks
            Hedgehog := Hedgehog.Next;
         end if;
      end loop;
      --hedgehog walks trough cycle, wife makes the same number of steps
      while Hedgehog.next /= Hare loop
         Hedgehog := Hedgehog.Next;
         Wife     := Wife.Next;
      end loop;
      --case: the whole list is the cycle (see README.TXT)
      if Wife.next = L then
         Wife.next := null;
      else
         Wife := Wife.next;
      end if;
      --hedgehog starts from first list element again
      Hedgehog := L;
      --wife finds last cycle element by finding hedgehog
      loop
         Hedgehog := Hedgehog.Next;
         if Wife.next = Hedgehog then
            --no more cycle:
            Wife.Next := null;
            --return pointer to first list element
            Hare:=L;
            return Hare;
         else
            Wife := Wife.Next;
         end if;
      end loop;   
   end;
   -------------------------------------------------       

   -- sample list
   My_List, Last, Iter: List := null;
   -- number of items in the list
   Num: List_Index;
   -- random number generation
   package Content_Random is new Ada.Numerics.Discrete_Random(Content); use Content_Random;
   Content_Gen: Content_Random.Generator;
   package Index_Random is new Ada.Numerics.Discrete_Random(List_Index); use Index_Random;
   Index_Gen: Index_Random.Generator;
   package Yes_No_Random is new Ada.Numerics.Discrete_Random(Boolean); use Yes_No_Random;
   Yes_No_Gen: Yes_No_Random.Generator;
begin
   -- initialize random number generators
   Reset(Content_Gen);
   Reset(Index_Gen);
   Reset(Yes_No_Gen);
   -- get the number of list items
   Num := Random(Index_Gen);
   Put("Erstelle eine zufaellig gefuellte Liste mit "); Put(Natural(Num), 0); Put_Line(" Elementen...");
   My_List := new List_Item'(Random(Content_Gen), null);
   Last := My_List;
   for I in 1..(Num - 1) loop
      My_List := new List_Item'(Random(Content_Gen), My_List);
   end loop;
   -- decide whether to create a cyle
   if Random(Yes_No_Gen) then
      declare
         type Cycle is new Natural range 1..Natural(Num);
         package Cycle_Random is new Ada.Numerics.Discrete_Random(Cycle); use Cycle_Random;
         Cycle_Gen: Cycle_Random.Generator;
         My_Cycle: Cycle;
      begin
         Reset(Cycle_Gen);
         My_Cycle := Random(Cycle_Gen);
         Put("Erstelle einen Zyklus zum Listenelement #"); Put(Natural(My_Cycle), 0); Put_Line("...");
         Iter := My_List;
         for I in 1..My_Cycle loop
            Iter := Iter.Next;
         end loop;
         Last.Next := Iter;
      end;
   else
      Put_Line("Erstelle keinen Zyklus.");
   end if;
   New_Line;
   -- call the function
   Put("Gibt es einen Zyklus? Die Funktion sagt: ");
   if Has_Cycle(My_List) then
      Put_Line("JA!");
   else
      Put_Line("NEIN!");
   end if;
   -------------------------------------------------
   --b)
   -------------------------------------------------
   if Has_Cycle(My_List) then
      Put_line("Entferne den Zyklus.");
      My_list := No_more_Cycle(My_list);
      -- did NO_more_Cycle work?
      Put("Gibt es immernoch einen Zyklus? Die Funktion sagt: ");
      if Has_Cycle(My_List) then
         Put_Line("JA!");
      else
         Put_Line("NEIN!");
      end if;
   end if;
end;
