--Author: Thomas Baumann, Steffen Keller, Andreas Krker
--Aufgabe: Packungsalgorithmen
with Text_Io, Ada.Float_Text_Io, Ada.Numerics.Discrete_Random;
use Text_Io, Ada.Float_Text_Io;

procedure Packung is

   type Behaelter is array (positive range <>) of Natural;
   
   type Ref_Behaelter is access Behaelter;
   
   --Globale Variable die die Kapazitaet der Behaelter festlegt
   Kapazitaet : Natural := 100;
   
   Eimer : Ref_Behaelter;
   Gegenstaende : Ref_Behaelter;
   Back, First, Best : Natural := 0;
   Avg : Float := 0.0;
   Sortieren : Boolean;
   Char : Character := '0';
   
   
   
   --Kapazitaetentyp fuer das Zufallszahlenpackage
   type Kap is new Integer Range 1 .. Kapazitaet;
   package Zufall is new Ada.Numerics.Discrete_Random(Kap);
   use Zufall;
   
   Seed : Generator;
   
   function Backtracking (Eimer : Ref_Behaelter; 
                          Gegenstaende : Ref_Behaelter;
                          Bereits_Eingefuegt : Natural;
                          Belegte_Behaelter : Natural;
                          Bestes_Ergebniss : Natural) return Natural;
   
   --Diese Funktion berechnet die beste Behaelter Anzahl per Backtracking
   --Dabei wird rekursiv vorgegangen. Zuerst wird der erste Gegenstand eingefuegt
   --Danach werden fuer jeden weiteren Gegenstand alle moeglichkeiten per
   --Tiefensuche ausprobiert. Dabei gilt:
   --der 1. Gegenstand wird nur im 1. Behalter versucht
   --der 2. Gegenstand wird nur in Behaelter 1 + 2 versucht
   --der 3. Gegenstand wird nur in 1 + 2 + 3 versucht
   --
   --Der Algorithmus wurde fast 1:1 aus dem Skript zur Vorlesung uebernommen.
   function Backtracking (Eimer : Ref_Behaelter; 
                          Gegenstaende : Ref_Behaelter;
                          Bereits_Eingefuegt : Natural;
                          Belegte_Behaelter : Natural;
                          Bestes_Ergebniss : Natural) return Natural 
         is
      Eim : Ref_Behaelter := Eimer;
      Geg : Ref_Behaelter := Gegenstaende;
      Used : Natural := Bereits_Eingefuegt;
      Belegt : Natural := Belegte_Behaelter;
      Best : Natural := Bestes_Ergebniss;
   begin
      --Fall der Algorithmus keine bessere als die bereits bestimme Loesung
      --funden kann wird die beste Loesung zurueck gegeben
      if Eimer(Bestes_Ergebniss) /= 0 then
         return Bestes_Ergebniss;
      --Wurden alle Gegenstaende eingefuegt geben wir die Anzahl
      elsif Used = Geg'Last then
         return Belegt;
      else
         --Hier werden die moeglichen, bereits belegten Behaelter ausprobiert
         for i in 1 .. Belegt loop
            if Eim(i) + Geg(Used + 1) <= Kapazitaet then
               Eim(i) := Eim(i) + Geg(Used + 1);
               Best := Backtracking(Eim, Geg, Used + 1, Belegt, Best);
               Eim(i) := Eim(i) - Geg(Used + 1);
            end if;
         end loop;
         --Ist noch ein Behalter frei, wird dieser Ebenfalls mit dem aktuellem
         --Gegenstand belegt
         if Belegt < Eim'Last then
            Eim(Belegt + 1) := Geg(Used + 1);
            --Diese Abfrage ist notwendig, da der Algorithmus sonst falsche
            --Ergebnisse liefern kann, wenn einer der Gegenstaende die Groesse
            --0 hat
            if Geg(Used + 1) = 0 then
               Best := Backtracking(Eim, Geg, Used + 1, Belegt, Best);
            else
               Best := Backtracking(Eim, Geg, Used + 1, Belegt + 1, Best);
            end if;
            Eim(Belegt + 1) := 0;
         end if;
      end if;
      
      --Rueckgabe des besten Ergebnisses
      return Best;
         
   end Backtracking;
   


   --Diese Funktion implementiert den First_Hit Algorithmus
   function First_Hit (Eimer : Ref_Behaelter; 
                       Gegenstaende : Ref_Behaelter) return Natural
         is
      Eim : Ref_Behaelter := new Behaelter (Eimer'Range);
      Geg : Ref_Behaelter := Gegenstaende;
      j : Integer;
   begin
      --Leeren der Eimer
      for I in 1 .. Eimer'Last loop
         Eim(I) := 0;
      end loop;
      
      --Jeder Gegenstand wird in den ersten leeren Behaelter gepackt
      for I in 1 .. Geg'Last loop
         --Zaehlvariable um den Behalter zu speichern in den gepackt wird
         J := 1;
         while Eim(J) + Geg(I) > Kapazitaet loop
            J := J + 1;
         end loop;
         Eim(J) := Eim(J) + Geg(I);
      end loop;
      
      --Der letzte Behaelter mit Inhalt wird gescuht und seine Nummer zurueck gegeben
      for i in 1 .. Eim'Last loop
         if Eim(i) = 0 then
            return i - 1;
         end if;
      end loop;
      
      return Eim'Last;
   end First_Hit;
   


   --Diese Funktion implemtiert den Best_Hit Algorithmus
   function Best_Hit (Eimer : Ref_Behaelter;
                      Gegenstaende : Ref_Behaelter) return Natural
         is
      Eim : Ref_Behaelter := new Behaelter(Eimer'Range);
      Geg : Ref_Behaelter := Gegenstaende;
      Min : Natural := 101;
      Beh_Nr : Natural := 1;
   begin
      --Leeren der Eimer
      for i in Eim'Range loop
         Eim(i) := 0;
      end loop;
      
      --Hier werden die Gegenstaende in die Behaelter eingefuegt
      for i in Geg'Range loop
         --Mit Hilfe von Min und Beh_Nr wird der Behaelter gespeichert,
         --in dem nach hinzufuegen des i'ten Gegenstandes der wenigste Platz
         --vorhanden ist
         --Min wird auf 101 gesetzt, damit fuer den Fall Geg(i) = 0 der erste
         --Eimer genommen wird
         Min := 101;
         Beh_Nr := 1;
         --Suchen des 'besten' Behalters
         for j in Eim'Range loop
            if Eim(j) + Geg(i) <= Kapazitaet and 100 - Eim(j) - Geg(i) < Min then
               Min := 100 - Eim(j) - Geg(i);
               Beh_Nr := j;
            end if;
         end loop;
         Eim(Beh_Nr) := Eim(Beh_Nr) + Geg(i);
      end loop;
      
      --Suchen des letzten vollen Behaelter, dessen Nummer zurueck gegeben wird
      for i in 1 .. Eim'Last loop
         if Eim(i) = 0 then
            return i - 1;
         end if;
      end loop;
      
      --Sind alle Behaelter gefuellt geben wir die Nr des letzten zurueck
      return Eim'Last;
   end Best_Hit;
   

   --Diese Funktion sortiert die Elemente des Ubergebenen Behaleters 
   --(in diesem Fall sind das die Gegenstaende) in Aufsteigender Reihenfolge
   --
   --Die Funktion wurde von Aufgabenblatt 1 - Aufgabe 6 uebernommen
   --nur das hier das sortierte Feld zurueckgegeben wird, anstadt das auf
   --einem globalen Fald operiert wird
   function Sort_Aufsteigend (Feld : Ref_Behaelter) return Ref_Behaelter
         is
      F : Ref_Behaelter := Feld;
      Weiter : Boolean := True;
      T : Natural := 0;
   begin
      while Weiter loop
         Weiter := False;
         for i in 1 .. F'Last - 1 loop
            if F(i) > F(i+1) then
               Weiter := True;
               T:=F(i);
               F(i) := F(i+1);
               F(i+1) := T;
            end if;
         end loop;
      end loop;
      
      return F;
   end Sort_Aufsteigend;
   


   --Diese Funktion sortiert die Elemente des Ubergebenen Behaleters 
   --(in diesem Fall sind das die Gegenstaende) in Absteigender Reihenfolge
   --
   --Die Funktion wurde von Aufgabenblatt 1 - Aufgabe 6 uebernommen,
   --nur das hier das sortierte Feld zurueckgegeben wird, anstadt das auf
   --einem globalen Fald operiert wird
   function Sort_Absteigend (Feld : Ref_Behaelter) return Ref_Behaelter
         is
      F : Ref_Behaelter := Feld;
      Weiter : Boolean := True;
      T : Natural := 0;
   begin
      while Weiter loop
         Weiter := False;
         for i in 1 .. F'Last - 1 loop
            if F(i) < F(i+1) then
               Weiter := True;
               T:=F(i);
               F(i) := F(i+1);
               F(i+1) := T;
            end if;
         end loop;
      end loop;
      
      return F;
   end Sort_Absteigend;

   
      


begin
   --User Interface
   --Dabei kann der benutzer waehlen, ob er die Gegenstaende sortieren moechte
   --und ob sie Auf- oder Absteigend sortiert werden sollen
   Put("Willkommen zum Testprogramm der 'Eintuetungsalgorithmen'!");
   New_Line;
   Put("Das Programm testen die Algorithmen mit 12 Gegenstaenden und 1000 Testlaeufen.");
   New_Line(2);
   While Char /= 'j' and Char /= 'n' loop
      Put("Moechten sie sortierte Felder testen(j/n)? ");
      Get(Char);
   end loop;
   
   Sortieren := Char = 'j';
   if Sortieren then
      while Char /= '1' and Char /= '2' loop
         Put("Moechten sie aufsteigend (1) oder absteigend (2) sortieren? ");
         Get(Char);
      end loop;
   end if;
      
   New_Line;
   
   --Initialisieren der Zufallszahlen
   Reset(Seed);
   
   --Erzeugen der Gegenstaende
   Eimer := new Behaelter(1 .. 12);
   Gegenstaende := new Behaelter(1 .. 12);
   
   --'leeren' der Behaelter
   for i in 1 .. 12 loop
      Eimer(i) := 0;
   end loop;
   
   --Wir fuehren 1000 Testlaeufe durch
   for i in 1 .. 1000 loop
      --Hier werden die 12 Gegenstaende zufaellig mit Zahlen 0 <= i <= 100 belegt
      for i in 1 .. 12 loop
         Gegenstaende(i) := Integer(Random(Seed));
      end loop;
      --Falls der Benutzer es wuenscht, werden die Gegenstaende sortiert
      if Sortieren then
         if Char = '1' then
            Gegenstaende := Sort_Aufsteigend(Gegenstaende);
         else
            Gegenstaende := Sort_Absteigend(Gegenstaende);
         end if;
      end if;
      --Back, First und Best speichern die gesammtzahl aller verwendeeter
      --Behaelter ueber alle 1000 Testlaeufe
      Back := Back + Backtracking(Eimer, Gegenstaende, 0, 0, 12);
      First := First + First_Hit(Eimer, Gegenstaende);
      Best := Best + Best_Hit(Eimer, Gegenstaende); 
   end loop;
   
   
   New_Line;
   Put("Die Durchscnittswerte fuer 12 Gegenstande und 1000 Testlaeufe: ");
   
   --Ausgabe der Durchschnittswerte fuer die Menge der Behaelter
   --Dabei wird alles nach der 4. Stelle abgeschnitten
   New_Line;
   Avg := Float(Back) / 1000.0;
   Put("Backtracking: "); Put(Avg, 2, 4, 0); Put(" Behaelter");
   New_Line;
   
   Avg := Float(First) / 1000.0;
   Put("First_Hit   : "); Put(Avg, 2, 4, 0); Put(" Behaelter");
   New_Line;
   
   Avg := Float(Best) / 1000.0;
   Put("Best_Hit    : "); Put(Avg, 2, 4, 0); Put(" Behaelter");
   New_Line;
end;

